//package mobicomp.layer;
package server;

import java.util.LinkedList;

/**
 * Represents a pool of reusable threads. Since creating and starting
 * a new thread is rather expensive, but we want to keep the thread
 * idea so that each socket layer can perform the way it needs to
 * without worrying about higher levels, this is a good compromise.
 */
public class ThreadPool {

	/** the max size */
	private int limit;

	/** the actual "pool" */
	private LinkedList pool;

	/** the type of threads */
	private Task task;

	private boolean debug = System.getProperty("server.debug.threads", "false").trim().equals("true");

	// debugging
	private ThreadGroup group = new ThreadGroup("mobicomp threads");

	/**
	 * Creates a new thread pool.
	 *
	 * @param class the {@link java.lang.Class) of {@link
	 * mobicomp.layer.ReusableThread}s that will be managed by this
	 * pool. Used for instantiating the threads. Needs to have an
	 * empty constructor.
	 * @param size the initial size of the pool
	 * @param limit the maximum number of threads that we allow in
	 * this pool
	 */
	public ThreadPool(Task task, int size, int limit) {
		if (limit > size) {
			this.limit = limit;
		} else {
			this.limit = size;
		}

		this.task = task;
		pool = new LinkedList();
		for (int i = 0; i < size; i++) {
			ReusableThread thread = new ReusableThread(task, this);
			pool.add(thread);
			(new Thread(thread, "reusable thread #" + i)).start();
		}
	}

	/**
	 * Puts the given thread back into the pool for reuse if the
	 * current size does not exceed the limit.
	 * @return <code>true</code> if the ReusableThread should keep
	 * running, <code>false</code> if it should exit
	 */
	synchronized boolean putBack(ReusableThread thread) {
		if (pool.size() < limit) {
			pool.add(thread);
			return true;
		} else {
			if (debug)
				System.err.println("Capacity reached. Exiting thread.");
			return false;
		}
	}

	/**
	 * Gets a thread out of the pool for usage.
	 * @return <code>null</code> if for some reason there is none
	 * available, otherwise a valid, reusable thread
	 */
	synchronized ReusableThread getThread() {
		ReusableThread thread = null;

		if (pool.size() == 0) {
			if (debug)
				System.err.println("ThreadPool overload. Starting new thread.");

			thread = new ReusableThread(task, this);
			thread.setThreadPool(this);
			(new Thread(group, thread, "additional thread")).start();
		} else {
			thread = (ReusableThread)pool.removeFirst();
		}

		if (debug)
			System.err.println("Additional threads: " + group.activeCount());

		return thread;
	}

}
